@@ -142,13 +142,12 @@ impl Not for Choice {
 ///
 /// Uses inline asm when available, otherwise it's a no-op.
 #[cfg(all(feature = "nightly", not(any(target_arch = "asmjs", target_arch = "wasm32"))))]
-fn black_box(input: u8) -> u8 {
+#[inline(always)]
+fn black_box(mut input: u8) -> u8 {
     debug_assert!((input == 0u8) | (input == 1u8));
 
-    // Pretend to access a register containing the input.  We "volatile" here
-    // because some optimisers treat assembly templates without output operands
-    // as "volatile" while others do not.
-    unsafe { asm!("" :: "r"(&input) :: "volatile") }
+    // Move value through assembler, which is opaque to the compiler, even though we don't do anything.
+    unsafe { asm!("" : "=r"(input) : "0"(input) ) }
 
     input
 }
@@ -156,10 +155,23 @@ fn black_box(input: u8) -> u8 {
 #[inline(never)]
 fn black_box(input: u8) -> u8 {
     debug_assert!((input == 0u8) | (input == 1u8));
-    // We don't have access to inline assembly or test::black_box or ...
+    // We don't have access to inline assembly or test::black_box, so we use the fact that
+    // volatile values will never be elided to register values.
     //
-    // Bailing out, hopefully the compiler doesn't use the fact that `input` is 0 or 1.
-    input
+    // Note: Rust's notion of "volatile" is subject to change over time. While this code may break
+    // in a non-destructive way in the future, it is better than doing nothing.
+
+    unsafe {
+        // Optimization barrier
+        //
+        // Unsafe is ok, because:
+        //   - &input is not NULL;
+        //   - size of input is not zero;
+        //   - u8 is neither Sync, nor Send;
+        //   - u8 is Copy, so input is always live;
+        //   - u8 type is always properly aligned.
+        core::ptr::read_volatile(&input as *const u8)
+    }
 }
 
 impl From<u8> for Choice {
